; Tool to branch to either FreeDOS or SYSLINUX, depending on whether
; you have a 386 (or newer) CPU or not. Based on metaboot by
; Eric Auer.CPUBOOT is Public Domain by Eric Auer, written in 2005.
; Purpose: http://syslinux.zytor.com/ SYSLINUX cannot boot - not even
; the boot sector) on pre-386 systems, so CPUBOOT can be used for
; automatic fallback to FreeDOS, to create universal boot disks.

; Compile: nasm -o cpuboot.bin cpuboot.asm
; Use: 1. SYSLINUX A: A:\SYSLINUX.BIN
;      copy files and create suitable syslinux boot sector
; 2. SYS A: A: /K CPUBOOT.SYS /BOOTONLY
;      install modified FreeDOS boot sector, with cpuboot.sys
;      instead of kernel.sys as name of the to be loaded file.
; 3. COPY /B cpuboot.bin + a:\syslinux.bin a:\cpuboot.sys
;      install cpuboot "kernel" with syslinux boot sector as
;      the one to chain to if a 386 or newer CPU is found.
; 4. copy a FreeDOS kernel to a:\kernel.sys - CPUBOOT will
;      re-run the FreeDOS boot sector and tell it to load
;      kernel.sys (if the CPU is too old for SYSLINUX).

        org 0	; flat binary file, loaded to 60:0, entry point 60:0.

	; called with boot drive in register BL. chains to a boot
	; sector which it loads to 0:7c00, entry point 0:7c00.
	; CPUBOOT calls the boot sector with DH=0x20,  DL=drive.
	; Chain to FreeDOS needs a FreeDOS boot sector still in RAM.

; *************************************************************

start:	mov dl,bl	; drive
	mov dh,20h	; bit 5: int 13 device
	push dx		; SAVE DRIVE

;---------------

	cld
	mov ax,0x1fe0	; FreeDOS boot sector should still be there
	mov ds,ax
	mov si,0x7c00	; 1fe0:7c00, not 1fe0:0...!
	xor ax,ax
	mov es,ax	; target: 0:7c00, normal boot sector location
	mov di,0x7c00
	mov cx,0x100
	rep movsw	; copy FreeDOS boot sector back to 0:7c00
	push cs
	pop ds		; reset ds - keep 0 in es

;---------------

	; test if we have 386 or better: 
	pushf		; save flags
	xor ax,ax
	push ax
	popf		; try to clear all bits
	pushf
	pop ax
	and ax,0f000h
	cmp ax,0f000h
	jz no386	; 4 msb stuck to 1: 808x or 80186
	mov ax,0f000h
	push ax
	popf		; try to set 4 msb
	pushf
	pop ax
	test ax,0f000h
	jz no386	; 4 msb stuck to 0: 80286

;---------------

is386:			; YES, a 386. We can chain to SYSLINUX.
	popf		; restore flags (one)

	mov si,noconcatmsg
	cmp [cs:0x3fe], word 0xaa55	; boot sector magic?
	jnz badluck	; would not be able to chain to SYSLINUX.

	cld
	mov si,0x200	; SYSLINUX boot sector is appended at 60:200
	mov di,0x7c00	; ES is still 0 - copy to 0:7c00
	mov cx,0x100
	rep movsw	; copy appended SYSLINUX boot sector to 0:7c00
	jmp short quietluck

;---------------

no386:			; no 386 found, so we must chain to FreeDOS
	popf		; restore flags (other)

	cmp [es:0x7dfe], word 0xaa55	; boot sector magic?
	jnz badluck	; give up otherwise

	mov di, ourname	; our own name!
	mov bp, 0x7c00	; search in boot sector
	call stringfind	; search in ES:BP..ES:BP+0x1f8, string DI, len 11
	jnz badluck	; Z = hit, at offset SI, DI unchanged

namefound:
	mov di, si	; overwrite the found string
	mov si, newname	; by "KERNEL  SYS".
	mov cx, 11
	rep movsb

goodluck:
	mov si,readytogomsg
	call print	; last message
quietluck:
	pop dx		; RESTORE DRIVE
	jmp word 0:0x7c00	; go for it! (and pass on DH and DL)

;---------------

badluck:		; tried to chain back to FreeDOS but failed
	mov si, abortmsg
	call print
	xor ax,ax
	int 0x16	; wait for a key
	int 0x19	; reboot warm
	mov al,0xfe
	out 0x64,al	; reboot cold

;---------------

stringfind:	; search for string at DI, len 11, in area ES:BP..ES:BP+0x1ff
			; return Z for hit at offset SI, DI unchanged
	push ax
	push cx
	xor ax,ax	; start search at buffer start
restringfind:
	mov si,bp	; buffer start
	add si,ax	; current search start
	mov cx,11
	push di
	push si		; needed for "return hit offset"
	xchg si,di	; we want ES for the buffer
	repe cmpsb	; compare strings here
	pop si		; needed for "return hit offset"
	pop di
	jz foundstring	; ZF set - found
	inc ax		; next search start
	cmp ax,0x200-11	; still in range?
	jb restringfind
	or ax,ax	; clear ZF - not found
foundstring:
	pop cx
	pop ax
	ret

; *************************************************************

print:			; print zero-terminated string at CS:SI
	push ax		; updates SI to point to the end of the string
	push ds
	mov ax,cs
	mov ds,ax
print2:	lodsb
	or al,al
	jz eprint
	call printchar
	jmp print2	; loop
eprint:	pop ds
	pop ax
	ret

;---------------

printchar:		; print char from AL
	push ax
	push bx
	mov ah,0x0e	; TTY
	xor bx,bx
	int 0x10	; BIOS screen stuff
	pop bx
	pop ax
	ret

; *************************************************************

ourname		db "CPUBOOT SYS"	; before patching

readytogomsg	db 13,10
		db "SYSLINUX needs at least a 386 CPU.",13,10
		db "Booting FreeDOS ("
newname		db "KERNEL  SYS" 	; after patching
readytogo2	db ") instead.",13,10,0

abortmsg	db 13,10
		db "Can neither boot SYSLINUX (PC too old)",13,10
		db "nor FreeDOS (unsupported boot loader?).",13,10
		db "Giving up. Press a key to reboot.",13,10,0

noconcatmsg	db 13,10
		db "Error: No boot sector at end of cpuboot.sys!",13,10,0

; *************************************************************

		; make us exactly 1 sector in size:
                times   0x1fc-$+$$ db 0
sign		db 0, 0, 0x55, 0xaa	; boot sector magic value

